@@ -1,15 +1,17 @@
 import asyncio
-from datetime import datetime
-from enum import Enum
 import functools
 import inspect
 import logging
+from copy import copy
+from datetime import datetime
+from enum import Enum
 from uuid import uuid4
 
 from tukio.dag import DAG
-from tukio.task import TaskTemplate, TaskRegistry, UnknownTaskName
-from tukio.utils import FutureState, Listen
-from tukio.broker import get_broker
+from tukio.task import TaskTemplate, TaskRegistry, UnknownTaskName, TukioTask
+from tukio.utils import FutureState, Listen, SkipTask
+from tukio.broker import get_broker, workflow_exec_topics
+from tukio.event import Event, EventSource
 
 
 log = logging.getLogger(__name__)
@@ -22,27 +24,47 @@ class WorkflowError(Exception):
 class WorkflowRootTaskError(WorkflowError):
 
     def __init__(self, value):
+        super().__init__()
         self._value = value
 
     def __str__(self):
         return 'expected one root task, found {}'.format(self._value)
 
 
-class WorkflowNotFoundError(WorkflowError):
-
-    def __str__(self):
-        return 'no workfow bound to the task, cannot unlock it!'
-
-
-class TemplateGraphError(Exception):
+class TemplateGraphError(WorkflowError):
 
     def __init__(self, key):
+        super().__init__()
         self._key = key
 
     def __str__(self):
         return 'graph error on task id: {}'.format(self._key)
 
 
+class RescueError(WorkflowError):
+
+    def __init__(self, uid, reason):
+        super().__init__()
+        self._uid = uid
+        self._reason = reason
+
+    def __str__(self):
+        return (
+            'cannot rescue workflow id {} from its last execution report. '
+            'reason: {}'.format(self._uid, self._reason)
+        )
+
+
+class WorkflowExecState(Enum):
+
+    begin = 'workflow-begin'
+    end = 'workflow-end'
+    error = 'workflow-error'
+    progress = 'workflow-progress'
+    suspend = 'workflow-suspend'
+    resume = 'workflow-resume'
+
+
 class OverrunPolicy(Enum):
 
     """
@@ -282,11 +304,11 @@ class WorkflowTemplate:
         template object.
         """
         wf_dict = {
-            "id": self.uid,
-            "policy": self.policy.value,
-            "topics": self.topics,
-            "tasks": [],
-            "graph": {}
+            'id': self.uid,
+            'policy': self.policy.value,
+            'topics': self.topics,
+            'tasks': [],
+            'graph': {}
         }
         for task_tmpl in self.tasks:
             wf_dict['tasks'].append(task_tmpl.as_dict())
@@ -306,8 +328,8 @@ class WorkflowTemplate:
     def validate(self):
         """
         Validate the current workflow template. At that point, we already know
-        the underlying DAG is valid. This methods ensures there's a single
-        root task and all task names are registered tasks.
+        the underlying DAG is valid. This method ensures there's a single root
+        task and all task names are registered tasks.
         If not valid, this method should raise either `WorkflowRootTaskError`
         or `UnknownTaskName` exceptions.
         """
@@ -315,24 +337,62 @@ class WorkflowTemplate:
         if root_nodes != 1:
             raise WorkflowRootTaskError(root_nodes)
         for task in self.tasks:
-            klass, _ = TaskRegistry.get(task.name)
-            # Check there's a `data_received` callback if the task is
-            # configured receive data during execution.
-            if (task.listen is not Listen.nothing and
-               not hasattr(klass, 'data_received')):
-                # Warn the user but don't raise an exception otherwise this
-                # would force the user to explicitely configure each
-                # coroutine-based tasks with {"topics": []}
-                log.warning("CAUTION: task '%s' has no callback to receive "
-                            "data during execution, will be ignored!",
-                            task.name)
+            TaskRegistry.get(task.name)
         return True
 
     def __str__(self):
         """
         Human readable string representation of a workflow template.
         """
-        return "<WorkflowTemplate uid={}>".format(self.uid)
+        return '<WorkflowTemplate uid={}>'.format(self.uid)
+
+
+def _current_workflow(func):
+    """
+    A decorator to maintain the dictionary of currently running workflows.
+    Inspired from the implementation of `asyncio.Task`.
+    """
+    @functools.wraps(func)
+    def wrapper(self, *args, **kwargs):
+        self.__class__._current_workflows[self._loop] = self
+        try:
+            return func(self, *args, **kwargs)
+        finally:
+            self.__class__._current_workflows.pop(self._loop)
+    return wrapper
+
+
+def _get_workflow_from_task(task):
+    """
+    Looks for an instance of `Workflow` linked to the task or a method of
+    `Workflow` among the done callbacks of the asyncio task. Returns None if
+    not found.
+    If the task was triggered from within a workflow it MUST have a `workflow`
+    attribute that points to it and at least one done callback that is a method
+    of `Workflow`.
+    """
+    if isinstance(task, TukioTask):
+        workflow = task.workflow
+    if workflow:
+        return workflow
+
+    for cb in task._callbacks:
+        # inspect.getcallargs() gives access to the implicit 'self' arg of
+        # the bound method but it is marked as deprecated since
+        # Python 3.5.1 and the new `inspect.Signature` object does NOT do
+        # the job :((
+        if inspect.ismethod(cb):
+            inst = cb.__self__
+        elif isinstance(cb, functools.partial):
+            try:
+                inst = cb.func.__self__
+            except AttributeError:
+                continue
+        else:
+            continue
+        if isinstance(inst, Workflow):
+            return inst
+    return None
 
 
 class Workflow(asyncio.Future):
@@ -342,6 +402,25 @@ class Workflow(asyncio.Future):
     way of workflow execution.
     """
 
+    # Inspired from the implementation of `asyncio.Task`
+    _current_workflows = {}
+
+    @classmethod
+    def current_workflow(cls, loop=None):
+        """
+        Returns the currently running workflow in an event loop or None.
+        By default the current workflow for the current event loop is returned.
+        None is returned when called not in the context of a Workflow.
+        """
+        loop = loop or asyncio.get_event_loop()
+        task = asyncio.Task.current_task(loop)
+        workflow = None
+        if task:
+            workflow = _get_workflow_from_task(task)
+        if not workflow:
+            workflow = cls._current_workflows.get(loop)
+        return workflow
+
     def __init__(self, wf_tmpl, *, loop=None, broker=None):
         super().__init__(loop=loop)
         self.uid = str(uuid4())
@@ -366,6 +445,13 @@ class Workflow(asyncio.Future):
             self.lock._locked = True
         # Work with an event broker
         self._broker = broker or get_broker(self._loop)
+        self._source = EventSource(
+            workflow_template_id=self._template.uid,
+            workflow_exec_id=self.uid
+        )
+        # A 'committed' workflow is a pending workflow not suspended
+        self._committed = asyncio.Event()
+        self._committed.set()
 
     @property
     def template(self):
@@ -375,7 +461,12 @@ class Workflow(asyncio.Future):
     def policy(self):
         return self._template.policy
 
-    def unlock(self, _):
+    @property
+    def committed(self):
+        return self._committed.is_set()
+
+    @_current_workflow
+    def _unlock(self, _):
         """
         The concept of 'locked workflow' only applies when the overrun policy
         is set to 'skip-until-unlock'. Once the workflow is unlocked, a new
@@ -384,6 +475,14 @@ class Workflow(asyncio.Future):
         if self.lock.locked():
             self.lock.release()
 
+    def unlock_when_task_done(self, task=None):
+        """
+        Adds a done callback to the passed or current task so as to unlock the
+        workflow when the task gets done.
+        """
+        task = task or asyncio.Task.current_task()
+        task.add_done_callback(self._unlock)
+
     def _register_to_broker(self, task_tmpl, task):
         """
         Registers the `data_received` callback of the task (if defined) into
@@ -394,14 +493,12 @@ class Workflow(asyncio.Future):
         if listen is Listen.nothing:
             return
 
-        # Try to register a callback, otherwise returns.
-        # Note: a warning message must have warned the user on loading the
-        # template in the engine.
+        # A tukio task always has a 'data_received' method
         try:
-            callback = task.holder.data_received
+            callback = task.data_received
         except AttributeError as exc:
-            log.debug('no callback to register in broker (%s), ignored', exc)
-            return
+            task.set_exception(exc)
+            raise
 
         # Register the callback in the event broker
         if listen is Listen.everything:
@@ -415,6 +512,7 @@ class Workflow(asyncio.Future):
                                     topics=task_tmpl.topics)
         task.add_done_callback(done_cb)
 
+    @_current_workflow
     def _unregister_from_broker(self, callback, _, topics=None):
         """
         A very simple wrapper around `Broker.unregister()` to ignore the future
@@ -433,75 +531,67 @@ class Workflow(asyncio.Future):
         if self._internal_exc:
             self._cancel_all_tasks()
 
-    def run(self, data):
+    @_current_workflow
+    def run(self, event):
         """
         Execute the workflow following the description passed at init.
         """
         # A workflow can be ran only once
         if self.tasks:
-            return
+            raise RuntimeError('a workflow can be run only once!')
+
         # Run the root task
-        root_tmpl = self._template.root()
-        task = self._new_task(root_tmpl, data)
-        self._start = datetime.utcnow()
-        # The workflow may fail to start at once
-        if not task:
+        try:
+            root_tmpl = self._template.root()
+        except WorkflowRootTaskError as exc:
+            self._internal_exc = exc
             self._try_mark_done()
+            task = None
+        else:
+            # Automatically wrap input data into an event object
+            self._dispatch_exec_event(WorkflowExecState.begin, copy(event))
+            task = self._new_task(root_tmpl, event)
+            self._start = datetime.utcnow()
+            # The workflow may fail to start at once
+            if not task:
+                self._try_mark_done()
         return task
 
-    def _new_task(self, task_tmpl, data):
+    def _new_task(self, task_tmpl, event):
         """
         Each new task must be created successfully, else the whole workflow
         shall stop running (wrong workflow config or bug).
+        It is assumed all tasks are tukio tasks (aka `TukioTask` objects).
         """
         try:
-            task = task_tmpl.new_task(data, loop=self._loop)
+            task = task_tmpl.new_task(event, loop=self._loop)
             # Register the `data_received` callback (if required) as soon as
             # the execution of the task is scheduled.
             self._register_to_broker(task_tmpl, task)
-        except (UnknownTaskName, TypeError, ValueError) as exc:
+        except (UnknownTaskName, TypeError, ValueError, AttributeError) as exc:
             log.error('failed to create task from template: %s', task_tmpl)
             self._internal_exc = exc
             self._cancel_all_tasks()
             return None
-        else:
-            log.debug('new task created for %s', task_tmpl)
-            done_cb = functools.partial(self._run_next_tasks, task_tmpl)
-            task.add_done_callback(done_cb)
-            self.tasks.add(task)
-            # Create the exec dict of the task
-            exec_dict = {'start': datetime.utcnow()}
-            try:
-                exec_dict['id'] = task.uid
-            except AttributeError:
-                exec_dict['id'] = None
-            self._tasks_by_id[task_tmpl.uid] = (task, exec_dict)
-            return task
+        task._workflow = self
+        log.debug('new task created for %s', task_tmpl)
 
-    def _join_task(self, task, result):
+        def next_tasks(future):
+            asyncio.ensure_future(self._run_next_tasks(future))
+
+        task.add_done_callback(next_tasks)
+        self.tasks.add(task)
+        # Create the exec dict of the task
+        self._tasks_by_id[task_tmpl.uid] = task
+        return task
+
+    def _join_task(self, next_task, event):
         """
-        Pass data to a downstream task that has already been started (by
+        Pass an event to a downstream task that has already been started (by
         another parent). In such a situation, it is known to be a join task.
         """
-        try:
-            # `data_received()` must be a simple callback (not a coroutine)
-            task.holder.data_received(result, from_parent=True)
-        except AttributeError as exc:
-            self._internal_exc = exc
-            self._cancel_all_tasks()
-            return False
-        return True
-
-    def _set_next_task_templates(self, task, next_tmpl_ids):
-        """
-        Defines the set of downstream tasks that should be executed right after
-        `task`. This method is intended to be called at runtime by the task
-        itself. `next_tmpl_ids` must be a list (can be empty) of task template
-        IDs.
-        It may be used to dynamically update (at runtime) the tasks that will
-        be executed by the workflow after `task`.
-        """
-        self._updated_next_tasks[task] = next_tmpl_ids
+        # Push event into next_task's event queue
+        asyncio.ensure_future(next_task.data_received(event))
 
     def _get_next_task_templates(self, task_tmpl, task):
         """
@@ -526,7 +616,21 @@ class Workflow(asyncio.Future):
         log.debug('%s filtered next tasks to: %s', task, filtered_tmpls)
         return filtered_tmpls
 
-    def _run_next_tasks(self, task_tmpl, task):
+    def _dispatch_exec_event(self, etype, data=None):
+        """
+        A shorthand to dispatch information about the workflow execution along
+        the way.
+        """
+        if isinstance(data, Event):
+            data = data.data
+        self._broker.dispatch(
+            {'type': etype.value, 'content': data},
+            topics=workflow_exec_topics(self.uid),
+            source=self._source,
+        )
+
+    @_current_workflow
+    async def _run_next_tasks(self, task):
         """
         A callback to be added to each task in order to select and schedule
         asynchronously downstream tasks once the parent task is done.
@@ -535,30 +639,59 @@ class Workflow(asyncio.Future):
         if self._must_cancel:
             self._try_mark_done()
             return
+
         # Don't execute downstream tasks if the task's result is an exception
         # (may include task cancellation) but don't stop executing the other
         # branches of the workflow.
         try:
             result = task.result()
-        except Exception as exc:
-            log.warning('task %s ended on exception', task_tmpl)
-            log.exception(exc)
-        else:
-            next_tmpls = self._get_next_task_templates(task_tmpl, task)
-            for tmpl in next_tmpls:
-                next_task, _ = self._tasks_by_id.get(tmpl.uid, (None, {}))
-                # Downstream task already running, join it!
-                if next_task:
-                    joined = self._join_task(next_task, result)
-                    if not joined:
-                        break
-                # Create new task
-                else:
-                    next_task = self._new_task(tmpl, result)
-                if not next_task:
-                    break
-        finally:
+        except SkipTask:
+            # If the task has been skipped, we just forward the previous task
+            # inputs to the next one.
+            result = task.inputs
+            log.warning('task %s has been skipped', task.template)
+        except asyncio.CancelledError:
+            log.warning('task %s has been cancelled', task.template)
             self._try_mark_done()
+            return
+        except Exception as exc:
+            log.warning('task %s ended on exception', task.template)
+            log.exception(exc)
+            self._try_mark_done()
+            return
+
+        # Ensure the workflow is not suspended, else wait for a resume.
+        try:
+            await self._committed.wait()
+        except asyncio.CancelledError:
+            # If cancelled, the workflow will return with the statement below.
+            log.debug('Suspended workflow cancelled')
+            return
+
+        source = EventSource(
+            workflow_template_id=self.template.uid,
+            workflow_exec_id=self.uid,
+            task_template_id=task.template.uid,
+            task_exec_id=task.uid
+        )
+        # Go through each child task
+        for tmpl in self._get_next_task_templates(task.template, task):
+            # Wrap result from parent task into an event object
+            event = Event(result, source=source)
+            next_task = self._tasks_by_id.get(tmpl.uid)
+            if next_task:
+                # Ignore done tasks
+                if next_task.done():
+                    continue
+                # Downstream task already running, join it!
+                self._join_task(next_task, event)
+            # Create new task
+            else:
+                next_task = self._new_task(tmpl, event)
+            if not next_task:
+                break
+
+        self._try_mark_done()
 
     def _try_mark_done(self):
         """
@@ -566,17 +699,27 @@ class Workflow(asyncio.Future):
         result is set to either the exec graph (represented by a dict) or to an
         exception raised at task creation.
         """
+        if not self._committed.is_set():
+            # Won't mark a workflow 'done' if it's suspended
+            return
+
         # Note: the result of the workflow may already have been set by another
         # done callback from another task executed in the same iteration of the
         # event loop.
         if self._all_tasks_done() and not self.done():
+            exec_event = WorkflowExecState.end
             if self._internal_exc:
                 self.set_exception(self._internal_exc)
+                exec_event = WorkflowExecState.error
+                data = self._internal_exc
             elif self._must_cancel:
                 super().cancel()
+                data = {'cancel': True}
             else:
                 self.set_result(self.tasks)
+                data = None
             self._end = datetime.utcnow()
+            self._dispatch_exec_event(exec_event, data=data)
 
     def _all_tasks_done(self):
         """
@@ -600,16 +743,77 @@ class Workflow(asyncio.Future):
             is_cancelled = task.cancel()
             if is_cancelled:
                 cancelled += 1
+                try:
+                    asyncio.ensure_future(task.holder.teardown())
+                except AttributeError:
+                    pass
         return cancelled
 
+    def set_next_tasks(self, task_tmpl_ids):
+        """
+        By default the workflow runs all downstream tasks once the current task
+        is done. This method allows to select the tasks that will be actually
+        ran and disables other tasks (unless they're already running).
+
+        This method is intended to be called at runtime by the task itself.
+        `task_tmpl_ids` must be a list (can be empty) of task template IDs.
+        """
+        task = asyncio.Task.current_task(self._loop)
+        if task not in self.tasks:
+            raise RuntimeError('task {} not executed by {}'.format(task, self))
+        self._updated_next_tasks[task] = task_tmpl_ids
+
+    def suspend(self):
+        """
+        Suspend the execution of this workflow prevents any new task from
+        starting.
+        """
+        if not self._committed.is_set():
+            return
+        self._committed.clear()
+
+        # Suspend a task means the task will be cancelled with the state
+        # 'suspended', which can be used later to resume the workflow
+        for task in (self.tasks - self._done_tasks):
+            task.suspend()
+
+        self._dispatch_exec_event(WorkflowExecState.suspend)
+        log.info('workflow %s has been suspended', self)
+
+    def resume(self):
+        """
+        Resume the execution of this workflow allows new tasks to start.
+        """
+        if self._committed.is_set():
+            log.error('Cannot resume a workflow that has not been suspended')
+            return
+        self._committed.set()
+
+        # Next tasks are done waiting for '_committed' asyncio event
+        # The 'suspended' ones needs to be re-executed
+        for task in self._done_tasks:
+            if FutureState.get(task) is not FutureState.suspended:
+                continue
+            event = Event(task.inputs, source=task.event_source)
+            self._new_task(task.template, event)
+
+        self._dispatch_exec_event(WorkflowExecState.resume)
+        log.info('workflow %s has been resumed', self)
+
     def cancel(self):
         """
         Cancel the workflow by cancelling all pending tasks (aka all tasks not
         marked as done). We must wait for all tasks to be actually done before
         marking the workflow as cancelled (hence done).
         """
+        if not self._committed.is_set():
+            self._committed.set()
+            self._try_mark_done()
+
         cancelled = self._cancel_all_tasks()
         if cancelled == 0:
+            # Tasks where already all done.If not,
+            # the last task will end the workflow already set as _must_cancel
             super().cancel()
         return True
 
@@ -617,7 +821,7 @@ class Workflow(asyncio.Future):
         """
         Readable string representation of a workflow execution object.
         """
-        string = "<Workflow template.uid={}, uid={}, start={}, end={}>"
+        string = '<Workflow template.uid={}, uid={}, start={}, end={}>'
         return string.format(self._template.uid, self.uid,
                              self._start, self._end)
 
@@ -626,30 +830,106 @@ class Workflow(asyncio.Future):
         Creates and returns a complete execution report, including workflow and
         tasks templates and execution details.
         """
-        wf_exec = {"id": self.uid, "start": self._start, "end": self._end}
-        wf_exec['state'] = FutureState.get(self).value
         report = self._template.as_dict()
-        report['exec'] = wf_exec
+        report['exec'] = {
+            'id': self.uid,
+            'start': self._start,
+            'end': self._end,
+            'state': FutureState.get(self).value
+        }
         # Update task descriptions to add info about their execution.
-        tasks_list = report['tasks']
-        for task_dict in tasks_list:
+        for task_dict in report['tasks']:
             try:
-                task, exec_dict = self._tasks_by_id[task_dict['id']]
+                task = self._tasks_by_id[task_dict['id']]
             except KeyError:
-                task_dict.update({'exec': None})
+                task_dict['exec'] = None
                 continue
-            exec_dict['state'] = FutureState.get(task).value
+            task_dict['exec'] = task.as_dict()
             # If the task is linked to a task holder, try to use its own report
             try:
-                task_report = task.holder.report()
+                task_dict['exec']['reporting'] = task.holder.report()
             except AttributeError:
                 pass
-            else:
-                if task_report is not None:
-                    exec_dict.update(task_report)
-            task_dict.update({'exec': exec_dict})
         return report
 
+    def fast_forward(self, report):
+        """
+        Restore a workflow and its tasks states, parsing an execution report.
+        """
+        self.uid = report['exec']['id']
+        self._start = report['exec']['start']
+        self._end = report['exec'].get('end')
+        self._source = EventSource(
+            workflow_template_id=self._template.uid,
+            workflow_exec_id=self.uid
+        )
+        log.info('fast forward workflow %s', self)
+
+        # Build a list of task in the graph that need to be executed
+        resume = list()
+
+        def browse(entry, parent=None):
+            """
+            Dive into the dag to:
+                - ensure executed task contexts are restored.
+                - find tasks that need to be executed.
+            """
+            t_template = next(
+                t for t in report['tasks'] if t['id'] == entry.uid
+            )
+            t_report = t_template.get('exec')
+            if not t_report:
+                if not parent:
+                    raise RescueError(self.uid, 'root task never been started')
+                # No execution report found, the task needs to be executed
+                resume.append((entry, None, parent))
+                return
+
+            t_state = FutureState(t_report['state'])
+            if not t_state.done():
+                # Pending, suspended or cancelled tasks need to be executed
+                resume.append((entry, t_template, parent))
+                return
+
+            if t_state.done():
+                # We won't manually create an asyncio task since it won't be
+                # run (the task is 'done'). Yet, we need a memory print of its
+                # execution (mainly for reporting purposes).
+                t_shadow = type('TukioTaskShadow', (object,), {
+                    'done': lambda: True,
+                    'cancelled': lambda: False,
+                    '_exception': None,
+                    'as_dict': lambda: t_report
+                })
+
+                # Add this task to the tracking sets
+                self.tasks.add(t_shadow)
+                self._done_tasks.add(t_shadow)
+                self._tasks_by_id[entry.uid] = t_shadow
+
+                # Recursive browing
+                for t_next in self._template.dag.successors(entry):
+                    browse(t_next, t_template)
+        browse(self._template.root())
+
+        # Resume/start tasks that need to be executed
+        for task_template, task_report, parent_report in resume:
+            if parent_report:
+                inputs = parent_report['exec']['outputs']
+                source = EventSource(
+                    workflow_template_id=self.template.uid,
+                    workflow_exec_id=self.uid,
+                    task_template_id=parent_report['id'],
+                    task_exec_id=parent_report['exec']['id']
+                )
+                event = Event(inputs, source=source)
+            else:
+                # No parent report = root task
+                event = Event(task_report['exec']['inputs'])
+            self._new_task(task_template, event)
+
+        self._dispatch_exec_event(WorkflowExecState.begin)
+
 
 def new_workflow(wf_tmpl, running=None, loop=None):
     """
@@ -659,63 +939,3 @@ def new_workflow(wf_tmpl, running=None, loop=None):
     """
     policy_handler = OverrunPolicyHandler(wf_tmpl, loop=loop)
     return policy_handler.new_workflow(running)
-
-
-class WorkflowInterface:
-
-    """
-    An object intended to be used from within tasks at runtime. It provides
-    methods to interact with the `Workflow` instance that is controlling the
-    task.
-    """
-
-    def __init__(self, task):
-        self.task = task
-        self._workflow = self._get_workflow(self.task)
-
-    def _get_workflow(self, task):
-        """
-        Looks for an instance of `Workflow` among the done callbacks of the
-        asyncio task. Raises a `WorkflowNotFoundError` if not found.
-        If the task was triggered from within a workflow it MUST have at least
-        one done callback that references the workflow itself.
-        """
-        for cb in task._callbacks:
-            # inspect.getcallargs() gives access to the implicit 'self' arg of
-            # the bound method but it is marked as deprecated since
-            # Python 3.5.1 and the new `inspect.Signature` object does NOT do
-            # the job :((
-            if inspect.ismethod(cb):
-                inst = cb.__self__
-            elif isinstance(cb, functools.partial):
-                try:
-                    inst = cb.func.__self__
-                except AttributeError:
-                    continue
-            else:
-                continue
-            if isinstance(inst, Workflow):
-                return inst
-        raise WorkflowNotFoundError
-
-    @classmethod
-    def from_current_task(cls):
-        """
-        Returns an instance of `WorkflowInterface` from the current task.
-        """
-        return cls(asyncio.Task.current_task())
-
-    def unlock_when_done(self):
-        """
-        Adds a done callback to the current task so as to unlock the workflow
-        that handles its execution when the task gets done.
-        """
-        self.task.add_done_callback(self._workflow.unlock)
-
-    def set_next_tasks(self, task_ids):
-        """
-        By default the workflow runs all downstream tasks once the current task
-        is done. This method allows to select the tasks that will be actually
-        ran and disables other tasks (unless they're already running).
-        """
-        self._workflow._set_next_task_templates(self.task, task_ids)
